//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000.                    //
//  All rights reserved.                                             //
//                                                                   //   
//  The database read by JFC is dirived directly from KANJIDIC       //
//  database dirived by Jim Breen.  Please see the _cpright.txt file //
//  for additional information.                                      //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------
//
//  This modlue implements various informational routines.  The 
//  principle of these are the character-info (kanji info) and the 
//  kanji-table.  These routines primrally provide information 
//  related to Japanese characters.
//
//  Structure of the KANJINFO.DAT files:
//
//  The following indicates the structure of the files written by this
//  utility.  The actual infirmaton content of the files can differ, but
//  the structure is identical.
//
//      MAGIC    -- 4 byte binary number used to verify the file type.
//      FLAGS    -- Flags indicating what is written into the file:
//      COUNT    -- Number of entries written into the file.
//      MAXJIS   -- Highest JIS code included in the file.
//
//                  KIFLAG_PINYIN (0x0001) -- Has pin yin data.
//                  KIFLAG_KOREAN (0x0002) -- Has Korean data.
//                  KIFLAG_NANORI (0x0004) -- Has nanori data.
//                  KIFLAG_EXTRA  (0x0008) -- Has extended data filed
//                  KIFLAG_EVAR   (0x0010) -- Has extended variable data
//                  KIFLAG_XREF   (0x0020) -- Has cross refference data
//
//      FIXED    -- A fixed file part.  This contains a sequency of 
//                  6355 kinfo structures representing each kanji.
//                  The structual fields are:
//                    
//                  offset      -- Points the VARIABLE part for this kanij.
//                  imi,nan,    -- Indicate the number entries of each
//                    on,kun,      type contained in the variable part.
//                    pinyin,      As data is removed to make th efile 
//                    korean,      smaller, these values are reduced (or
//                    extra        set to zero) and the correspoding 
//                                 entries are removed from the file.
//                                          
//      VARIABLE -- The variable part consists of a number of null 
//                  termianted strings writtent sequecially into the 
//                  file.  The above counters indicate the number 
//                  of each type of string.  The order of the strings
//                  is as follows:
//
//                  Korean   -- Korean data if pressent.
//                  PinYin   -- Ascii with extended charactrs to get
//                              accents.
//                  meanings -- Ascii.
//                  on-yomi  -- Compressed katakana.  The first byte is 
//                              removed, to save space.  The special
//                              code 0x1f indicates a katakana -.
//                  kun-yomi -- Compressed hiragana.  The first byte
//                              is removed, to save space.  The special
//                              code 0x1f indicates a katakana -, and 
//                              the special code 0x20 indicates the 
//                              beginning of the okurigana.
//                  nanori   -- Compressed hiragana.  Special cdoes 
//                              above are supported but not used.
//                  Extra    -- Extended field if pressent (see below for description)
//
//      EXTRA -- The extra is a binary field.  This must be the last field, because it
//               is possible to get embbedded zeros in this field.  The layout of the 
//               field is a fixed length binary structrue followed by a sequence of 
//               one-byte codes followed by the data for that argument.  The data format 
//               varies according to the idividual code field.  Much of this data 
//               is not pressent in all entries, so this type of coding takes less space.  
//               Additionally, this allows the design to change depending on what is to 
//               be added later.  
//
//              F<short>        -- Frequency count.
//              I<short>        -- Spahn and Hadamitzky Kana & Kanji
//              E<short>        -- "A Guide To Remembering Japanese Characters" by Kenneth G. Henshall
//              K<short>        -- Gakken Kanji Dictionary ("A  New Dictionary  of Kanji Usage")
//              L<short>        -- "Remembering The Kanji" by James Heisig
//              O<short>        -- "Japanese Names", by P.G. O'Neill
//              Q<short>        -- Four-corners secondary code <code><resolution is stroed in the struture>
//
//  The organization of the extra field, will always include all entries before the cross-reference
//  entries are added.  This allows the reader to mark the frist cross-reference entry and 
//  decode from there forward.  Note the cross-reference codes are all lower case, which makes
//  them easy to identifiy.
//
//              h<short>        -- Halpern dictionary cross-refernece entries.
//              n<short>        -- Nelson dictionary cross-reference entries.
//              o<short>        -- "Japanese Names", by P.G. O'Neill cross-refernece
//              k<short>        -- JIS 0208 Cross-reference
//              j<short>        -- JIS 0212 Cross-reference
//              z<short>        -- SKIP code missclassification/cross-reference 
//                                  Coded as (<error><<13) | (<type> << 10) | (<1> << 5)) | (<2>)
//
//      DROPPED ENTRY (only one of these):
//              i<short>        -- Spahn and Hadamitzky Kana & Kanji cross-reference 
//                                 coded as (rstorke << 11) | (ostroke<<6)
//
//  The first part of the file conists of a number of kinfo 
//  structures that are arranged simply in a packed order of the 
//  kanji in the JIS table.  Note that the structure is packed, and 
//  data elements are allingned on word boundaries.  This gives a total
//  structure size of 14 bytes, which is unusual for higher end systems.
//
//  The on/kun lines are stored in a compressed format, that is simply 
//  the last byte of the caracter.  Thus they must be combained with 
//  BASE_HIRAGANA, or BASE_KATAKANA to form an actual character.  Within
//  the on lines, the special character 0x1f is used to indicate a 
//  katakana -.  Within the kun lines, the special character 0x1f is 
//  used to indicate a katakana -, and the special character 0x20 is 
//  used to indicate the start of an okurigana.  This translates to 
//  a Japanese (.  Note that the closing ) is not contained within the 
//  info file, and must be added by the programmer.  (Note the 
//  implementation I use does not distinguish between the on/kun lists,
//  and uses the same interpreter for both.  Since, however, the on-yomi
//  should not contain the code 0x20, this is not a problem.)
//
//  Understanding the big buffer:
//
//  Because the information contained in the second part of the kanji-
//  info file is free form (i.e. varies in length), this information 
//  is broken up by having it in lines.  Unfortunately this presents 
//  some major problems.  First, windows file system does not deal 
//  with line delminated files (ther is no read-line).  Normally, one
//  might use the c++ or c file systems to read this type of data.  
//  In this case I do not want to load either of these systems, to keep
//  the code size small and fast.  
//
//  To get around this problem, I use brute force.  I read a big buffer
//  full of data starting at the location idnciated by the kinfo.
//  offset field and break this up in memory.  This has two plusses.
//  it is really fast, and it is really easy.
//
//  Most of this message is old, the buffer space required is small 
//  now that I am using my own database files.  The databese building
//  utility will actually tell you how long the buffer needs to be.
//

#include "jfc_main.h"
#include "jfc_conf.h"
#include "jfc_edit.h"
#include "jfc_font.h"
#include "jfc_help.h"
#include "jfc_info.h"
#include "jfc_jisc.h"
#include "jfc_misc.h"

//-------------------------------------------------------------------
//
//  Compile time options.
//
#define NAME_KANJIINFO  TEXT("kanjinfo.dat")    // Name of the main kanji-info file.

#define KINFO_MAGIC     0x34A5B4D4      // This value is used to confirm that 
                                        //   this is the correct verison of the 
                                        //   KANJINFO.DAT file.

#define KANJIINFO_OFFSET    (2*sizeof(ulong)+2*sizeof(ushort))  //  Offset of first entry in the file.

#define USE_ASCII_OKURIGANA             // If defined causes the system to 
                                        //   use ascii ( and ) to surround
                                        //   okurigana, instead of using 
                                        //   Japanese versions.  This looks
                                        //   better to me.

//#define FIRST_PAGE_ONLY                 // This cuases the system to display on the first page,
                                        //   This removes the More Info and Cross Reference  
                                        //   dialogs.

//-------------------------------------------------------------------
//
//  static data and definitions.
//
//  These are various static data used in some of the routines.
//

static KANJI_info *info;    // Global pointer so routines can find the class instance.

#define CHARTYPE_UNKNOWN    0       // Character types.
#define CHARTYPE_ASCII      1       // Needs to be here because these move in different
#define CHARTYPE_OEM        2       //   versions of the code.
#define CHARTYPE_JSYMBOL    3
#define CHARTYPE_JASCII     4
#define CHARTYPE_HIRAGANA   5
#define CHARTYPE_KATAKANA   6
#define CHARTYPE_GREEK      7
#define CHARTYPE_RUSSIAN    8
#define CHARTYPE_RESERVED   9
#define CHARTYPE_KANJI1     10
#define CHARTYPE_KANJI2     11

//-------------------------------------------------------------------
//
//  static routines.
//
static byte *skip_line   (byte *ptr);                                               // Skip to next entry in data pointer
static byte *put_entry   (KANJI *buffer,int base,byte *ptr);                        // Add a reading to a stirng for a card
static byte *put_line    (EUC_buffer *line,int base,byte *ptr,int add);             // Format an output line for the kanji-info list-box.
static byte *put_reading (EUC_buffer *line,int base,byte *ptr,int index,int last);  // Special rotuine used to place readings.

static BOOL CALLBACK dialog_kanjiinfo  (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam);
static BOOL CALLBACK dialog_moreinfo   (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam);
static BOOL CALLBACK dialog_xrefinfo   (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam);

//-------------------------------------------------------------------
//
//  Small service rotuines used to display a particular kind of kanji 
//  information data.
//

//
//  Displays four-corner data, and well as checking to see if the data
//  needs to be displayed. 
//
//      hwnd  -- Dialog box pointer.
//      id    -- ID for location to display.
//      main  -- Main part of the data (set to -1 to indicate non-data).
//      index -- Last digiti that resolveds the differences between diffrerent 
//               entrys with the same main value.
//
#if (!defined(FIRST_PAGE_ONLY) || !defined(WINCE))
static void info_fourcorner (HWND hwnd,int id,int main,int index) {
  TCHAR buffer[40];
  if (main == -1) return;
  wsprintf (buffer,TEXT("%04d.%d"),main,index);
  SetDlgItemText (hwnd,id,buffer);
  return;
}
#endif (defined(FIRST_PAGE_ONLY) && defined(WINCE))

//
//  Displays an ASCII coded string.  For CE machines, this must be converted 
//  to UNICODE before the display.  For NT/98/95 machines, this is simply 
//  a macro.
//
//      hwnd   -- Dialog box pointer.
//      id     -- Location to display the text.
//      string -- Data to display.
//
#ifdef WINCE
static void info_string (HWND hwnd,int id,byte *string) {
  TCHAR text[SIZE_BUFFER];
  MultiByteToWideChar (1252,0,(char *) string,-1,text,SIZE_BUFFER);
  SetDlgItemText      (hwnd,id,text);
  return;
}
#else WINCE
  #define info_string(w,i,s) SetDlgItemText(w,i,(char *) s);
#endif WINCE

//
//  This utility rotuine formats a string for the a flash card.  This is
//  very simular to the put_line() procedure, but the output is to a buffer,
//  not a list box.
//
//      line   -- Pointer to EUC_buffer class object associated with the 
//                list box.
//      base   -- Character to base to use.  This determines the type of 
//                line being formatted:
//
//          0             -- Meaning line.
//          BASE_KATAKANA -- on-yomi.
//          BASE_HIRAGANA -- kun-yomi.
//
//      ptr    -- Pointer to current location in buffer.  To optimize the 
//                file operations, all the line information in the knaji-
//                info file is read in one block and decoded in memory. 
//                On entry this pointer points to the beginning of the 
//                line.  The RETRUN value points the beginning of the 
//                next line.
//
//      RETURN -- Points to the beginning of the next line in the buffer.
//
//  The following special characters are supported:
//
//      0x1f (all bases)         -- produces a katakana -.
//      high-bit set (all bases) -- produces a kanji (, indicating the 
//                                  beginning of a origana.  The closing 
//                                  ) will be added by the formating 
//                                  routine automatically.
//
#ifdef USE_ASCII_OKURIGANA
  #define PARAN_LEFT    '('
  #define PARAN_RIGHT   ')'
#else 
  #define PARAN_LEFT    KANJI_LPARAN
  #define PARAN_RIGHT   KANJI_RPARAN
#endif USE_ASCII_OKURIGANA

static byte *put_entry (KANJI *buffer,int base,byte *ptr) {
  int okurigana = false;
  int adding    = false;
  if (*buffer) adding = true;                   // Are we adding to a buffer.
  while (*buffer) buffer++;                     // Find the end of the buffer
  if (adding) {                                 // If adding need a Japanese or ASCII cama
    if (base) *buffer++ = KANJI_COMMA;
      else {
        *buffer++ = ',';
        *buffer++ = ' ';
      }
  }
  while (*ptr) {                                // Build a line.
    if (*ptr == 0x1f) *buffer++ = KANJI_DASH;
      else {
        if (*ptr & 0x80) { *buffer++ = PARAN_LEFT; okurigana = true; }
        *buffer++ = (base | (*ptr & 0x7F));
      }
    ptr++;
  }
  if (okurigana) *buffer++ = PARAN_RIGHT;       // Main need a termiantion.
  *buffer = 0;                                  // Clean-up
  ptr++;
  return (ptr);
}

//
//  This utility rotuine formats a string for the list-box containned
//  in the kanji-info dialog box.  This routine is used for on-yomi, 
//  kun-yomi, meanings, and titles.
//
//      line   -- Pointer to EUC_buffer class object associated with the 
//                list box.
//      base   -- Character to base to use.  This determines the type of 
//                line being formatted:
//
//          0             -- Meaning line.
//          BASE_KATAKANA -- on-yomi.
//          BASE_HIRAGANA -- kun-yomi.
//
//      ptr    -- Pointer to current location in buffer.  To optimize the 
//                file operations, all the line information in the knaji-
//                info file is read in one block and decoded in memory. 
//                On entry this pointer points to the beginning of the 
//                line.  The RETRUN value points the beginning of the 
//                next line.
//      add    -- If set to true, causes this to be added to the current
//                line instead of repolacing the current line.
//
//      RETURN -- Points to the beginning of the next line in the buffer.
//
//  The following special characters are supported:
//
//      0x1f (all bases)         -- produces a katakana -.
//      high-bit set (all bases) -- produces a kanji (, indicating the 
//                                  beginning of a origana.  The closing 
//                                  ) will be added by the formating 
//                                  routine automatically.
//
#ifdef USE_ASCII_OKURIGANA
  #define PARAN_LEFT    '('
  #define PARAN_RIGHT   ')'
#else  USE_ASCII_OKURIGANA
  #define PARAN_LEFT    KANJI_LPARAN
  #define PARAN_RIGHT   KANJI_RPARAN
#endif USE_ASCII_OKURIGANA

static byte *put_line (EUC_buffer *line,int base,byte *ptr,int add) {
  int okurigana = false;
  if (!add) line->clear ();
  while (*ptr) {
    if (*ptr == 0x1f) line->put_char (KANJI_DASH);
      else {
        if (*ptr & 0x80) { line->put_char(PARAN_LEFT); okurigana = true; }
        line->put_char (base | (*ptr & 0x7F));
      }
    ptr++;
  }
  if (okurigana) line->put_char (PARAN_RIGHT);
  if (!add) line->flush (-1);
  return (ptr+1);
}

//
//  This is a wrapper for the rotuine put_line() that does special processing for 
//  putting the reading fields for kanji.   This routine handles the compressed/
//  non-compressed display of the character data.
//
//      line   -- Pointer to EUC_buffer class object associated with the 
//                list box.
//      base   -- Character to base to use.  This determines the type of 
//                line being formatted:
//
//          0             -- Meaning line.
//          BASE_KATAKANA -- on-yomi.
//          BASE_HIRAGANA -- kun-yomi.
//
//      ptr    -- Pointer to current location in buffer.  To optimize the 
//                file operations, all the line information in the knaji-
//                info file is read in one block and decoded in memory. 
//                On entry this pointer points to the beginning of the 
//                line.  The RETRUN value points the beginning of the 
//                next line.
//      index  -- Current index into the reading type being processed.
//      last   -- Last index in reading type being processed.
//
//      RETURN -- Points to the beginning of the next line in the buffer.
//
static byte *put_reading (EUC_buffer *line,int base,byte *ptr,int index,int last) {
  if (!jfc_config.cfg.info_compress) return (put_line(line,base,ptr,false));    // Uncompressed line, just pass through.
  if (!index) line->clear ();                       // Start of a compressed block.
    else {          
      if (base) line->put_char (KANJI_COMMA);       // Continuing a compressed block
        else {
          line->put_char (',');
          line->put_char (' ');
        }
    }
  ptr = put_line(line,base,ptr,true);               // Output actual line
  if (index+1 == last) line->flush (-1);            // End of compressed block, flush the converter
  return (ptr);
}

//
//  Small utility routine used to skip some strings in the line buffer
//  during processing.  Basically this is used to skip to the meaning 
//  strings first.
//
//      ptr    -- Pointer to current location in buffer.  To optimize the 
//                file operations, all the line information in the knaji-
//                info file is read in one block and decoded in memory. 
//                On entry this pointer points to the beginning of the 
//                line.  The RETRUN value points the beginning of the 
//                next line.
//
//      RETURN -- Points to the beginning of the next line in the buffer.
//
static byte *skip_line (byte *ptr) {
  ptr += strlen((char *) ptr)+1;
  return (ptr);
}

//-------------------------------------------------------------------
//
//  Dialog box stub routines.
//

//
//  Stub routine for kanji-info dialog box.
//
//  This routine has to save the parameter passed during the intialization.
//  This parameter is the pointer to the KANJI_info object associated
//  with the dialog.  The routine then always recalls the saved 
//  parameter and uses it to call the object's dialog box procedure.
//
static BOOL CALLBACK dialog_kanjiinfo (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam) {
  return (info->dlg_kanjiinfo(hwnd,message,wParam,lParam));
}

//
//  Dialog box procedure for the display window that pops up so you can
//  get a better look at the kanji in the PPC version.  This dialog box
//  should be created with a call to DialogBoxParam, so the kanji to 
//  be displayed in the dialog can be passed to the program!
//
#ifdef WINCE_PPC
static byte kanji_view = false; // This variable is used to keep this 
                                // dialog from being created more than
                                // once.  When the dialog is entered,
                                // this is set to true.  This will prevent
                                // the kanji view from launching the 
                                // dialog again.  When the dialog is 
                                // termianted, this is cleared.

static BOOL CALLBACK dialog_kanjiview (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam) {
  switch (message) {
    case WM_INITDIALOG:
         kanji_view = true;     // Block second creation.
         SendDlgItemMessage (hwnd,IDC_KIBIGKANJI,WMU_SETWINDOWVALUE,0,lParam);
         return (false);
//
//  Process help messages
//
    case WM_HELP:
         do_help (hwnd,IDH_KANJI_CHARINFO);
         return  (true);
//
//  Process push buttons.
//
    case WM_COMMAND:    
         switch (LOWORD(wParam)) {
           case IDOK:
           case IDCANCEL:
                kanji_view = false;     // Allow dialog to be generated again.
                EndDialog (hwnd,false);
                return (true);
         }
         break;
  }
  return (false);
}
#endif WINCE_PPC

//
//  Stub routine for the more-info dialog box.
//
//  This routine has to save the parameter passed during the intialization.
//  This parameter is the pointer to the KANJI_info object associated
//  with the dialog.  The routine then always recalls the saved 
//  parameter and uses it to call the object's dialog box procedure.
//
#ifndef FIRST_PAGE_ONLY
static BOOL CALLBACK dialog_moreinfo (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam) {
  return (info->dlg_moreinfo(hwnd,message,wParam,lParam));
}
#endif FIRST_PAGE_ONLY

//
//  Stub routine for the xref-info dialog box.
//
//  This routine has to save the parameter passed during the intialization.
//  This parameter is the pointer to the KANJI_info object associated
//  with the dialog.  The routine then always recalls the saved 
//  parameter and uses it to call the object's dialog box procedure.
//
#if (defined(WINCE_PPC) && !defined(FIRST_PAGE_ONLY))
static BOOL CALLBACK dialog_xrefinfo (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam) {
  return (info->dlg_xrefinfo(hwnd,message,wParam,lParam));
}
#endif (defined(WINCE_PPC) && !defined(FIRST_PAGE_ONLY))

//-------------------------------------------------------------------
//
//  Window procedures
//

//
//  Window procedure for window class used to display the bushu character
//  This class simply accepts a single bit of data that determins the 
//  bushu character.  Setting this character to zero will suppress display
//  of the bushu.
//
static KANJI bushu_symbols[] = {
    0x306c, 0x2143, 0x5026, 0x254e, 0x3235,     /* 01 - 05 */
    0x502d, 0x4673, 0x5035, 0x3f4d, 0x5139,     /* 06 - 10 */
    0x467e, 0x482c, 0x5144, 0x514c, 0x5152,     /* 11 - 15 */
    0x515c, 0x5161, 0x4561, 0x4e4f, 0x5231,     /* 16 - 20 */
    0x5238, 0x5239, 0x523e, 0x3d3d, 0x4b4e,     /* 21 - 25 */
    0x5247, 0x524c, 0x5253, 0x4b74, 0x387d,     /* 26 - 30 */
    0x5378, 0x455a, 0x3b4e, 0x5469, 0x546a,     /* 31 - 35 */
    0x4d3c, 0x4267, 0x3d77, 0x3b52, 0x555f,     /* 36 - 40 */
    0x4023, 0x3e2e, 0x5577, 0x5579, 0x5625,     /* 41 - 45 */
    0x3b33, 0x406e, 0x3929, 0x384a, 0x3652,     /* 46 - 50 */
    0x3433, 0x5676, 0x5678, 0x572e, 0x5730,     /* 51 - 55 */
    0x5735, 0x355d, 0x2568, 0x5744, 0x5746,     /* 56 - 60 */
    0x3f34, 0x5879, 0x384d, 0x3c6a, 0x3b59,     /* 61 - 65 */
    0x5a3d, 0x4a38, 0x454d, 0x3654, 0x4a7d,     /* 66 - 70 */
    0x5a5b, 0x467c, 0x5b29, 0x376e, 0x4c5a,     /* 71 - 75 */
    0x3767, 0x3b5f, 0x5d46, 0x5d55, 0x5d59,     /* 76 - 80 */
    0x4866, 0x4c53, 0x3b61, 0x5d63, 0x3f65,     /* 81 - 85 */
    0x3250, 0x445e, 0x4963, 0x602b, 0x602d,     /* 86 - 90 */
    0x4a52, 0x3267, 0x356d, 0x3824, 0x383c,     /* 91 - 95 */
    0x364c, 0x313b, 0x3424, 0x3445, 0x4038,     /* 96 - 100 */
    0x4d51, 0x4544, 0x4925, 0x4942, 0x6222,     /* 101 - 105 [ 104 questionable ] */
    0x4772, 0x4869, 0x3b2e, 0x4c5c, 0x4c37,     /* 106 - 110 */
    0x4c70, 0x4050, 0x3c28, 0x633b, 0x3253,     /* 111 - 115 [ 114 questionable ] */
    0x376a, 0x4e29, 0x435d, 0x4a46, 0x3b65,     /* 116 - 120 */
    0x344c, 0x6626, 0x4d53, 0x3129, 0x4f37,     /* 121 - 125 */
    0x3c29, 0x6650, 0x3c2a, 0x6666, 0x4679,     /* 126 - 130 */
    0x3f43, 0x3c2b, 0x3b6a, 0x3131, 0x4065,     /* 131 - 135 */
    0x4124, 0x3d2e, 0x3a31, 0x3f27, 0x6767,     /* 136 - 140 */
    0x6948, 0x436e, 0x376c, 0x3954, 0x3061,     /* 141 - 145 */
    0x403e, 0x382b, 0x3351, 0x3840, 0x432b,     /* 146 - 150 */
    0x4626, 0x6c35, 0x6c38, 0x332d, 0x4056,     /* 151 - 155 */
    0x4176, 0x422d, 0x3f48, 0x3c56, 0x3f49,     /* 156 - 160 */
    0x4324, 0x6d68, 0x4d38, 0x4653, 0x4850,     /* 161 - 165 [ 162 questionable ] */
    0x4e24, 0x3662, 0x4439, 0x4c67, 0x496c,     /* 166 - 170 */
    0x7030, 0x7032, 0x312b, 0x4044, 0x4873,     /* 171 - 175 */
    0x4c4c, 0x3357, 0x706a, 0x706c, 0x323b,     /* 176 - 180 */
    0x4a47, 0x4977, 0x4874, 0x3f29, 0x3c73,     /* 181 - 185 */
    0x3961, 0x474f, 0x397c, 0x3962, 0x7175,     /* 186 - 190 */
    0x7228, 0x722e, 0x722f, 0x3534, 0x357b,     /* 191 - 195 */
    0x443b, 0x7343, 0x3c2f, 0x734e, 0x4b63,     /* 196 - 200 */
    0x322b, 0x3550, 0x3975, 0x7363, 0x7366,     /* 201 - 205 */
    0x4524, 0x385d, 0x414d, 0x4921, 0x736e,     /* 206 - 210 */
    0x736f, 0x4e36, 0x737d, 0x737e,             /* 211 - 214 */
};

static LRESULT CALLBACK JFC_bushu_proc (HWND hwnd,UINT iMsg,WPARAM wParam,LPARAM lParam) {
  HDC           hdc;
  CREATESTRUCT *create;
  PAINTSTRUCT   ps;
  int           bushu;

  switch (iMsg) {
//
//  Here we primarally need to adjust the size of the window.
//
    case WM_CREATE:                                 
         create = (CREATESTRUCT *) lParam;
         MoveWindow (hwnd,create->x,create->y,sys_font.hwidth,sys_font.height,true);
         lParam = 0;
//
//  This is the set command that sets the color we are to display.
//
    case WMU_SETWINDOWVALUE:        // *** FALL THROUGH *** 
         SetWindowLong  (hwnd,0,lParam);
         InvalidateRect (hwnd,NULL,true);
         return (0);
//
//  This does the actual redraw.
//
    case WM_PAINT:
         hdc   = BeginPaint (hwnd,&ps);
         bushu = GetWindowLong(hwnd,0);
         if ((bushu >= 1) && (bushu <= 214)) {  
           sys_font.kanji->draw (hdc,bushu_symbols[bushu-1],0,sys_font.kanji->height);
         }
         EndPaint (hwnd,&ps);
         return (0);
  }
  return (DefWindowProc(hwnd,iMsg,wParam,lParam));
}

//
//  Window producedure for the control to draw the big kanji character.
//  This is a relativly simple rendering routine.
//
static LRESULT CALLBACK JFC_kanji_proc (HWND hwnd,UINT iMsg,WPARAM wParam,LPARAM lParam) {
  HDC           hdc;        // Display context
  CREATESTRUCT *create;     // Create window structure
  PAINTSTRUCT   ps;         // Update paint structure
  LOGFONT       lf;         // Logical-Font structure used to make large 
                            //   version of the ascii font for the big-char box.
  RECT          rect,full;  // Rectangle
  KANJI_font   *big;        // Large kanji font for window.
  HFONT         font;       // Font for ascii font.
  SIZE          size;       // Used to get the size of the ascii characters.
  TCHAR         text[2];    // Temp buffer to display ascii characters.
  int           kanji;      // The character

  switch (iMsg) {
//
//  Here we primarally need to adjust the size of the window.
//
    case WM_CREATE:                                 
         create = (CREATESTRUCT *) lParam;
         lParam = 0;
//
//  This is the set command that sets the color we are to display.
//
    case WMU_SETWINDOWVALUE:        // *** FALL THROUGH *** 
         SetWindowLong  (hwnd,0,lParam);
         InvalidateRect (hwnd,NULL,true);
         return (0);
//
//  For PPC version, we allow clicking on the small kanji view to 
//  make a large kanji view.
//
#ifdef WINCE_PPC
    case WM_LBUTTONDOWN:
         kanji = GetWindowLong(hwnd,0);
         if (kanji && !kanji_view) JDialogBox (IDD_KANJIVIEW,hwnd,(DLGPROC) dialog_kanjiview,kanji);
         return (0);
#endif WINCE_PPC
//
//  This does the actual redraw.
//
    case WM_PAINT:
         hdc   = BeginPaint (hwnd,&ps);
         kanji = GetWindowLong(hwnd,0);
         GetClientRect (hwnd,&full);
         rect.right  = full.right -sys_font.hspace;
         rect.left   = full.left  +sys_font.hspace;
         rect.top    = full.top   +sys_font.vspace;
         rect.bottom = full.bottom-sys_font.vspace;
         if (kanji) {  
//
//  Render kanji characters.
//
           if (ISJIS(kanji)) {
             big = get_bigfont(&full);
             big->fill (hdc,kanji,&rect);    // Kanji characters are easy!
           }
//
//  Ascii characters involve making a very big font, and fitting the 
//  font into the box.  If the PROCESS_WIDE_CHARCTERS is on, we may 
//  have to do the whole thing twice.  Ther first time, we generate 
//  the font, we see if the character is too wide.  If it is then we
//  propotinally scale down the font, and make a other attempt.
//
           else {
             memset (&lf,0,sizeof(lf));
             lstrcpy (lf.lfFaceName,jfc_config.cfg.ascii_font.name);
             lf.lfHeight = rect.top-rect.bottom;
             if (!(font = CreateFontIndirect(&lf))) return (true);
             font = (HFONT) SelectObject (hdc,font);
             text[0] = (TCHAR) kanji;
             text[1] = 0;
             SetBkMode (hdc,TRANSPARENT);
             GetTextExtentPoint32 (hdc,text,1,&size);
#ifdef PROCESS_WIDE_CHARACTERS
             if (size.cx > rect.right-rect.left) {    // Character is too wide.
               SelectObject (hdc,font);
               DeleteObject (font);
               lf.lfHeight = ((rect.top-rect.bottom)*(rect.right-rect.left))/size.cx;
               if (!(font = CreateFontIndirect (&lf))) return (true);
               font = (HFONT) SelectObject (hdc,font);
               GetTextExtentPoint32 (hdc,text,1,&size);
             }
#endif PROCESS_WIDE_CHARACTERS
             TextOut (hdc,rect.left+(rect.right-rect.left-size.cx)/2,rect.top,text,1);
             SelectObject (hdc,font);
             DeleteObject (font);
           }
         }
         EndPaint (hwnd,&ps);
         return (0);
  }
  return (DefWindowProc(hwnd,iMsg,wParam,lParam));
}

//-------------------------------------------------------------------
//
//  begin class KANJI_info
//
//  This class implements the kanji-info lookup.
//

static byte *info_cache = NULL;     // This location holds the cached kanji information. 

//
//  Generates the kanji info dialog box for a particular kanji.
//  
//      kanji -- Kanji (character) to be processed.
//
void KANJI_info::dialog (HWND hwnd,int kanji) {
  info = this;
  ch   = kanji; 
  JDialogBox (IDD_KANJIINFO,hwnd,dialog_kanjiinfo);
  SetFocus   (hwnd);
  return;
}

//
//  This is the main routine.  This basically is the dialog procedure
//  for the kanji-info dialog box.
//
int KANJI_info::dlg_kanjiinfo (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam) {
  switch (message) {
//
//  This is the main part of the routine.  Since this is just an 
//  informational dialog box, the main part is in making the dialog 
//  box.
//
    case WM_INITDIALOG: 
         init_dialog (hwnd);
         return      (false);
//
//  Process help messages
//
    case WM_HELP:
         do_help (hwnd,IDH_KANJI_CHARINFO);
         return  (true);
//
//  Process push buttons.
//
    case WM_COMMAND:    
         switch (LOWORD(wParam)) {
           case IDOK:
           case IDCANCEL:
                EndDialog (hwnd,false);
                return    (true);
#ifndef FIRST_PAGE_ONLY
           case IDC_KIMORE:
                JDialogBox (IDD_MOREINFO,hwnd,(DLGPROC) dialog_moreinfo);
                return (true);
#ifdef WINCE_PPC
           case IDC_KIXREF:
                JDialogBox (IDD_XREFINFO,hwnd,(DLGPROC) dialog_xrefinfo);
                return (true);
#endif WINCE_PPC
#endif FIRST_PAGE_ONLY
         }
         break;
  }
  return (false);
}

//
//  This is the dialog box handler for the More Kanji Info dialog.  There are
//  two copies of this routine.  One for PPC machines and one for all other 
//  routines (the gneral rotuine has some variation between CE and non-CE 
//  machines.
//
#ifndef FIRST_PAGE_ONLY
#ifdef  WINCE_PPC

//
//  Dialog box handler for PPC machines for the More Kanji Info dialog.
//
int KANJI_info::dlg_moreinfo (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam) {
  static short types[] = { IDS_KI_TYPEUNKNOWN,IDS_KI_TYPEASCII,IDS_KI_TYPEEXTENDED,IDS_KI_TYPESYMBOL,IDS_KI_TYPEJASCII,IDS_KI_TYPEHIRAGANA,IDS_KI_TYPEKATAKANA,IDS_KI_TYPEGREEK,IDS_KI_TYPERUSSIAN,IDS_KI_TYPERESERVED,IDS_KI_TYPEKANJI1,IDS_KI_TYPEKANJI2 };
  TCHAR buffer[SIZE_BUFFER];
  int   i;
  switch (message) {
//
//  This is the main part of the routine.  Since this is just an 
//  informational dialog box, the main part is in making the dialog 
//  box.
//
    case WM_INITDIALOG: 
         SetFocus (GetDlgItem(hwnd,IDC_MINEXT));
//
//  Do JIS value
// 
//
//  Utility routine to generate the character pattern geneated by varius
//  Japanese character encodings.  This is used to see what will show 
//  up when viewing a file.
//
         wsprintf (buffer,TEXT("%X"),jis2sjis(ch));
         SetDlgItemText (hwnd,IDC_KISHIFTJIS,buffer);
         wsprintf (buffer,TEXT("%X"),jis2unicode(ch));
         SetDlgItemText (hwnd,IDC_KIUNICODE,buffer);
//
//  Do character type.
//
         i = HIBYTE(ch);
         if       (ch == 0)                   i = CHARTYPE_UNKNOWN;
         else if  (ch <= 127)                 i = CHARTYPE_ASCII;
         else if  (ch <= 255)                 i = CHARTYPE_OEM;
         else if ((i == 0x21) || (i == 0x22)) i = CHARTYPE_JSYMBOL;
         else if  (i == 0x23)                 i = CHARTYPE_JASCII;
         else if  (i == 0x24)                 i = CHARTYPE_HIRAGANA;
         else if  (i == 0x25)                 i = CHARTYPE_KATAKANA;
         else if  (i == 0x26)                 i = CHARTYPE_GREEK;
         else if  (i == 0x27)                 i = CHARTYPE_RUSSIAN;
         else if ((i >= 0x28) && (i <= 0x2f)) i = CHARTYPE_RESERVED;
         else if ((i >= 0x30) && (i <= 0x4f)) i = CHARTYPE_KANJI1;
         else if ((i >= 0x50) && (i <= 0x74)) i = CHARTYPE_KANJI2;
         else                                 i = CHARTYPE_UNKNOWN;
         SetDlgItemText (hwnd,IDC_KITYPE,get_string(types[i]));
         if (i < CHARTYPE_KANJI1) return (false);   // Not kanji we have have no more info.
//
//  Process PinYin.
//
         if (korean) info_string (hwnd,IDC_KIKOREAN,korean);
         if (pinyin) info_string (hwnd,IDC_KIPINYIN,pinyin);
//
//  Process the entended data entries in the list.
//
         if (kinfo.extra) {
           info_fourcorner (hwnd,IDC_KIFC ,extend.fc_main,extend.fc_index);
           info_fourcorner (hwnd,IDC_KIFC2,fc_main2      ,extend.fc_index2);
           if (freq    ) SetDlgItemInt (hwnd,IDC_MIFREQ    ,freq    ,false);
           if (henshall) SetDlgItemInt (hwnd,IDC_MIHENSHALL,henshall,false);
           if (gakken  ) SetDlgItemInt (hwnd,IDC_MIGAKKEN  ,gakken  ,false);
           if (heisig  ) SetDlgItemInt (hwnd,IDC_MIHEISIG  ,heisig  ,false);
           if (oneill  ) SetDlgItemInt (hwnd,IDC_MIONEILL  ,oneill  ,false);
           if (extend.md_short1) {
             wsprintf (buffer,TEXT("%d.%04d"),extend.md_short1,extend.md_short2);
             SetDlgItemText (hwnd,IDC_KIMDSHORT,buffer);
           }
           if (extend.md_long) {
             if      (extend.md_x) wsprintf (buffer,TEXT("%dX"),extend.md_long);
             else if (extend.md_p) wsprintf (buffer,TEXT("%dP"),extend.md_long);
             else                  wsprintf (buffer,TEXT("%d") ,extend.md_long);
             SetDlgItemText (hwnd,IDC_KIMDLONG,buffer);
           }
         }
         return (false);
//
//  Process help messages
//
    case WM_HELP:
         do_help (hwnd,IDH_KANJI_CHARINFO);
         return  (true);
//
//  Process push buttons.
//
    case WM_COMMAND:    
         switch (LOWORD(wParam)) {
           case IDOK:
           case IDCANCEL:
                EndDialog (hwnd,false);
                return (true);
           case IDC_MINEXT:
                JDialogBox (IDD_XREFINFO,hwnd,(DLGPROC) dialog_xrefinfo,(LONG) this);
                return (true);
         }
         break;
  }
  return (false);
}

#else WINCE_PPC

//
//  General dialog box handler for the More Kanji Info dialog.
//
int KANJI_info::dlg_moreinfo (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam) {
  static short errors[] = { IDS_KI_SKIPPOSITION,IDS_KI_SKIPSTROKE,IDS_KI_SKIPBOTH,IDS_KI_SKIPBREEN };
  switch (message) {
//
//  This is the main part of the routine.  Since this is just an 
//  informational dialog box, the main part is in making the dialog 
//  box.
//
    case WM_INITDIALOG: 
#ifdef WINCE
         info_fourcorner (hwnd,IDC_KIFC ,extend.fc_main,extend.fc_index);
         info_fourcorner (hwnd,IDC_KIFC2,fc_main2      ,extend.fc_index2);
         if (korean) info_string (hwnd,IDC_KIKOREAN,korean);
         if (pinyin) info_string (hwnd,IDC_KIPINYIN,pinyin);
#endif WINCE
         if (kinfo.extra) {
           byte *ptr;
           int   i;
           TCHAR buffer[SIZE_BUFFER];
           EUC_buffer line;                     // EUC line buffer used to set strings in the list-box.
           if (freq    ) SetDlgItemInt (hwnd,IDC_MIFREQ    ,freq    ,false);
           if (henshall) SetDlgItemInt (hwnd,IDC_MIHENSHALL,henshall,false);
           if (gakken  ) SetDlgItemInt (hwnd,IDC_MIGAKKEN  ,gakken  ,false);
           if (heisig  ) SetDlgItemInt (hwnd,IDC_MIHEISIG  ,heisig  ,false);
           if (oneill  ) SetDlgItemInt (hwnd,IDC_MIONEILL  ,oneill  ,false);
           line.initialize (GetDlgItem(hwnd,IDC_MIXREF));   // Initialize buffer for output.
           for (ptr = xref; xref && *ptr; ptr += 3) {
             i = *((ushort *) (ptr+1));
             switch (*ptr) {
               case 'n':
                    format_string (buffer,IDS_MI_NELSON,i);
                    break;
               case 'h':
                    format_string (buffer,IDS_MI_HALPERN,i);
                    break;
               case 'o':
                    format_string (buffer,IDS_MI_ONEILL,i);
                    break;
               case 'k':
                    format_string (buffer,IDS_MI_JIS0208,i);
                    break;
               case 'j':
                    format_string (buffer,IDS_MI_JIS0212,i);
                    break;
               case 'z':
                    format_string (buffer,IDS_MI_SKIP,((i >> 10) & 0x7),((i >> 5) & 0x1f),(i & 0x1f),get_string(errors[(i >> 13)-1]));
                    break;
             }
#ifdef WINCE
             for (i = 0; buffer[i]; i++) ((byte *) buffer)[i] = (char) buffer[i]; 
             ((byte *) buffer)[i] = 0;
#endif WINCE        
             put_line (&line,0,(byte *) buffer,false);
           }
#ifdef WINCE
           if (extend.md_short1) {
             wsprintf (buffer,TEXT("%d.%04d"),extend.md_short1,extend.md_short2);
             SetDlgItemText (hwnd,IDC_KIMDSHORT,buffer);
           }
           if (extend.md_long) {
             if      (extend.md_x) wsprintf (buffer,TEXT("%dX"),extend.md_long);
             else if (extend.md_p) wsprintf (buffer,TEXT("%dP"),extend.md_long);
             else                  wsprintf (buffer,TEXT("%d") ,extend.md_long);
             SetDlgItemText (hwnd,IDC_KIMDLONG,buffer);
           }
#endif WINCE
         }
         SetFocus (GetDlgItem(hwnd,IDC_MIXREF));
         return (false);
//
//  Process help messages
//
    case WM_HELP:
         do_help (hwnd,IDH_KANJI_CHARINFO);
         return  (true);
//
//  Process push buttons.
//
    case WM_COMMAND:    
         switch (LOWORD(wParam)) {
           case IDOK:
           case IDCANCEL:
                EndDialog (hwnd,false);
                return (true);
         }
         break;
  }
  return (false);
}

#endif WINCE_PPC
#endif FIRST_PAGE_ONLY

//
//  This is the dialog box handler for the Cross-Reference dialog.  This dialog
//  is only used on PPC machines, and only because there is not sufficient 
//  room for these items in any of the other dailog pages.
//
#if (defined(WINCE_PPC) && !defined(FIRST_PAGE_ONLY))
int KANJI_info::dlg_xrefinfo (HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam) {
  static short errors[] = { IDS_KI_SKIPPOSITION,IDS_KI_SKIPSTROKE,IDS_KI_SKIPBOTH,IDS_KI_SKIPBREEN };
  switch (message) {
//
//  This is the main part of the routine.  Since this is just an 
//  informational dialog box, the main part is in making the dialog 
//  box.
//
    case WM_INITDIALOG: 
         if (kinfo.extra) {
           byte *ptr;
           int   i;
           TCHAR buffer[SIZE_BUFFER];
           EUC_buffer line;                     // EUC line buffer used to set strings in the list-box.
           line.initialize (GetDlgItem(hwnd,IDC_MIXREF));   // Initialize buffer for output.
           for (ptr = xref; xref && *ptr; ptr += 3) {
             i = *((ushort *) (ptr+1));
             switch (*ptr) {
               case 'n':
                    format_string (buffer,IDS_MI_NELSON,i);
                    break;
               case 'h':
                    format_string (buffer,IDS_MI_HALPERN,i);
                    break;
               case 'o':
                    format_string (buffer,IDS_MI_ONEILL,i);
                    break;
               case 'k':
                    format_string (buffer,IDS_MI_JIS0208,i);
                    break;
               case 'j':
                    format_string (buffer,IDS_MI_JIS0212,i);
                    break;
               case 'z':
                    format_string (buffer,IDS_MI_SKIP,((i >> 10) & 0x7),((i >> 5) & 0x1f),(i & 0x1f),get_string(errors[(i >> 13)-1]));
                    break;
             }
             for (i = 0; buffer[i]; i++) ((byte *) buffer)[i] = (char) buffer[i]; 
             ((byte *) buffer)[i] = 0;
             put_line (&line,0,(byte *) buffer,false);
           }
         }
         SetFocus (GetDlgItem(hwnd,IDC_MIXREF));
         return (false);
//
//  Process help messages
//
    case WM_HELP:
         do_help (hwnd,IDH_KANJI_CHARINFO);
         return  (true);
//
//  Process push buttons.
//
    case WM_COMMAND:    
         switch (LOWORD(wParam)) {
           case IDOK:
           case IDCANCEL:
                EndDialog (hwnd,false);
                return (true);
         }
         break;
  }
  return (false);
}
#endif (defined(WINCE_PPC) && !defined(FIRST_PAGE_ONLY))

//
//  Get the kanji-info for a particular kanji.
//
//      ch     -- Character to get info for.
//      full   -- Inidcates call all infromation possible, otherwise a minimal
//                set is returned.  The full set is used by the kanji information
//                feature, but the minimal set is used to make cards.
//
void KANJI_info::get_info (int ch,int full) {
  unsigned long done;
  int   i;
  byte *ptr;
//
//  Get fixed data elements.
//
  ch  = (ch & 0x7f7f)-0x3021;
  ch  = HIBYTE(ch)*94+LOBYTE(ch);
  if (ch > KIMAX_KANJI) {                       // User asked for an invalid character
    memset (&kinfo,0,sizeof(kinfo));
    return;
  }
  if (info_cache) memcpy (&kinfo,info_cache+KANJIINFO_OFFSET+ch*sizeof(kinfo),sizeof(kinfo));
    else {
      SetFilePointer (handle,KANJIINFO_OFFSET+ch*sizeof(kinfo),NULL,FILE_BEGIN);
      ReadFile (handle,&kinfo,sizeof(kinfo),&done,NULL);
    }
//
//  Get the basic data sets, meanings, etc.
//
  if (info_cache) ptr = info_cache+kinfo.offset;
    else {
      SetFilePointer (handle,kinfo.offset,NULL,FILE_BEGIN);
      ReadFile (handle,buffer,SIZE_INFOBUFFER,&done,NULL);
      ptr = buffer;
    }
  if (!kinfo.korean) korean = NULL; else ptr = skip_line(korean = ptr);
  if (!kinfo.pinyin) pinyin = NULL; else ptr = skip_line(pinyin = ptr);
  for (imi = ptr, i = 0; i < kinfo.imi; i++) ptr = skip_line(ptr);
  for (on  = ptr, i = 0; i < kinfo.on ; i++) ptr = skip_line(ptr);
  for (kun = ptr, i = 0; i < kinfo.kun; i++) ptr = skip_line(ptr);
  nan = ptr;
//
//  Exit for short read
//
  if (!full) return;
//
//  Process the extended data.
//
  if (!kinfo.extra) {
    memset (&extend,0,sizeof(struct extend));
    extend.fc_main = ~0UL;                                          // -1 in a portable way
  }
  else {
    for (i = 0; i < kinfo.nan; i++) ptr = skip_line(ptr);           // Skip the nanori
    memcpy (&extend,ptr,sizeof(struct extend));                     // Fixed part of the extended data.
    freq = sh_kana = henshall = gakken = heisig = oneill = 0;       // Zero out the data fields.
    fc_main2 = -1;
    xref     = NULL;
    for (ptr += sizeof(struct extend); *ptr; ptr += 3) {            // Process the data list.
      if (islower(*ptr)) {                                          // Lowercase entry indicates start
        xref = ptr;                                                 //   of the cross-reference entries
        break;                                                      //   processing of these is delayed
      }                                                             //   until later.
      i = (((ushort) ptr[2]) << 8) | ptr[1];                        // Get interger parameter.
      switch (*ptr) {                                               // Common entries 
        case 'F': freq     = i; break;
        case 'I': sh_kana  = i; break;
        case 'E': henshall = i; break;
        case 'K': gakken   = i; break;
        case 'L': heisig   = i; break;
        case 'O': oneill   = i; break;
        case 'Q': fc_main2 = i; break;
        default:
             break;
      }
    }
  }
  return;
}

//
//  This is the handler for the WM_INITDIALOG message for the main Kanji Info 
//  dialog.  This has been seperatred out from the dialog handler to allow for
//  easier addaptation between the different platforms.  There are two main 
//  versions (PPC and non-PPC) machines (the non-PPC version has some CE/non-CE
//  variation.
//
#ifdef WINCE_PPC
//
//  PPC handler for WM_INITDIALOG for the Kanji Info dialog
//
void KANJI_info::init_dialog (HWND hwnd) {
  TCHAR      buffer[SIZE_BUFFER];       // Text buffer for building lines.
  EUC_buffer line;                      // EUC line buffer used to set strings in the list-box.
  TCHAR      buffer2[20];               // Buffer for accumulating pin-yin data and ASCII value of JIS code
  byte      *ptr;                       // Pointer to current location within the data buffer.
  int        i;
  if (ch > 0x00ff) ch &= 0x7f7f;        // We accept EUC characters.
//
//  Blank out all old data.
//
  SetDlgItemText (hwnd,IDC_KIJISCODE,TEXT(""));
  for (i = IDC_KISTROKES; i <= IDC_KISHKANA; i++) SetDlgItemText (hwnd,i,TEXT(""));
  SendDlgItemMessage (hwnd,IDC_KILIST,JL_RESET,0,0);
//
//  Do JIS value
// 
//
//  Utility routine to generate the character pattern geneated by varius
//  Japanese character encodings.  This is used to see what will show 
//  up when viewing a file.
//
  if (ch <= 0x00ff) { buffer2[0] = (byte) ch; buffer2[1] = 0; }
    else { buffer2[0] = HIBYTE(ch); buffer2[1] = LOBYTE(ch); buffer2[2] = 0; }
  wsprintf (buffer,TEXT("%X (%X) [%s]"),ch,ch | 0x8080,buffer2);
  SetDlgItemText (hwnd,IDC_KIJISCODE,buffer);
  SendDlgItemMessage (hwnd,IDC_KIBIGKANJI     ,WMU_SETWINDOWVALUE,0,ch);
  SendDlgItemMessage (hwnd,IDC_KIBUSHUCHAR    ,WMU_SETWINDOWVALUE,0,0);
  SendDlgItemMessage (hwnd,IDC_KICLASSICALCHAR,WMU_SETWINDOWVALUE,0,0);
//
//  Do character type.
//
  i = HIBYTE(ch);
  if       (ch == 0)                   i = CHARTYPE_UNKNOWN;
  else if  (ch <= 127)                 i = CHARTYPE_ASCII;
  else if  (ch <= 255)                 i = CHARTYPE_OEM;
  else if ((i == 0x21) || (i == 0x22)) i = CHARTYPE_JSYMBOL;
  else if  (i == 0x23)                 i = CHARTYPE_JASCII;
  else if  (i == 0x24)                 i = CHARTYPE_HIRAGANA;
  else if  (i == 0x25)                 i = CHARTYPE_KATAKANA;
  else if  (i == 0x26)                 i = CHARTYPE_GREEK;
  else if  (i == 0x27)                 i = CHARTYPE_RUSSIAN;
  else if ((i >= 0x28) && (i <= 0x2f)) i = CHARTYPE_RESERVED;
  else if ((i >= 0x30) && (i <= 0x4f)) i = CHARTYPE_KANJI1;
  else if ((i >= 0x50) && (i <= 0x74)) i = CHARTYPE_KANJI2;
  else                                 i = CHARTYPE_UNKNOWN;
#ifndef FIRST_PAGE_ONLY
  EnableWindow(GetDlgItem(hwnd,IDC_KIMORE),i >= CHARTYPE_KANJI1);
#endif  FIRST_PAGE_ONLY
  if (i < CHARTYPE_KANJI1) return;          // Not kanji we have have no more info.
  memset (&kinfo,0,sizeof(kinfo));
  if (((ch & 0xff) < 0x21) || ((ch & 0xff) > 0x7e)) return;
//
//  Open and read kanji information, an display the easy items.
//        
  if (open_info()) return;
  get_info (ch,INFO_ALL);
  close_info ();
  SendDlgItemMessage (hwnd,IDC_KIBUSHUCHAR,WMU_SETWINDOWVALUE,0,kinfo.bushu);
  SetDlgItemInt (hwnd,IDC_KISTROKES,kinfo.strokes,true);
  if (!kinfo.classical) SetDlgItemInt (hwnd,IDC_KIBUSHU,kinfo.bushu,true);
    else {
      wsprintf (buffer2,TEXT("%d (%d)"),kinfo.bushu,kinfo.classical);
      SetDlgItemText     (hwnd,IDC_KIBUSHU,buffer2);
      SendDlgItemMessage (hwnd,IDC_KICLASSICALCHAR,WMU_SETWINDOWVALUE,0,kinfo.classical);
    }       
  if (kinfo.grade  ) SetDlgItemInt  (hwnd,IDC_KIGRADE  ,kinfo.grade  ,false);
  if (kinfo.nelson ) SetDlgItemInt  (hwnd,IDC_KINELSON ,kinfo.nelson ,false);
  if (kinfo.halpern) SetDlgItemInt  (hwnd,IDC_KIHALPERN,kinfo.halpern,false);
  if (kinfo.haig   ) SetDlgItemInt  (hwnd,IDC_KIHAIG   ,kinfo.haig   ,false);
  wsprintf (buffer2,TEXT("%d-%d-%d"),kinfo.skip_t,kinfo.skip_1,kinfo.skip_2);
  if (kinfo.skip_t ) SetDlgItemText (hwnd,IDC_KISKIP,buffer2);
//
//  Read all of the line type entries all in one big block.  This is 
//  very brut-force, but it does save time/space, and everything else.
//
  line.initialize (GetDlgItem(hwnd,IDC_KILIST));    // Initialize buffer for output.
//
//  Process list elements.
//
  if (jfc_config.cfg.info_titles) put_line (&line,0,(byte *) "\x01\x1fmeanings\x1f",false);
  for (ptr = imi, i = 0; i < kinfo.imi; i++) ptr = put_reading(&line,0,ptr,i,kinfo.imi);
  if (jfc_config.cfg.info_titles && kinfo.on ) put_line (&line,0,(byte *) "\x01\x1fon-yomi\x1f",false);
  for (ptr = on,  i = 0; i < kinfo.on; i++) ptr = put_reading(&line,BASE_KATAKANA,ptr,i,kinfo.on);
  if (jfc_config.cfg.info_titles && kinfo.kun) put_line (&line,0,(byte *) "\x01\x1fkun-yomi\x1f",false);
  for (ptr = kun, i = 0; i < kinfo.kun; i++) ptr = put_reading(&line,BASE_HIRAGANA,ptr,i,kinfo.kun);
  if (jfc_config.cfg.info_titles && kinfo.nan) put_line (&line,0,(byte *) "\x01\x1fnanori\x1f",false);
  for (ptr = nan, i = 0; i < kinfo.nan; i++) ptr = put_reading(&line,BASE_HIRAGANA,ptr,i,kinfo.nan);
//
//  Process the entended data entries in the list.
//
  if (kinfo.extra) {
    if (extend.sh_rstroke || extend.sh_ostroke) {
      wsprintf (buffer,TEXT("%d%c%d.%d"),extend.sh_rstroke,extend.sh_radical+'a',extend.sh_ostroke,extend.sh_index);
      SetDlgItemText (hwnd,IDC_KISHDICT,buffer);
    }
    if (sh_kana) SetDlgItemInt (hwnd,IDC_KISHKANA,sh_kana,false);
  }
//
//  Activate the list so the user can use the cursor keys to move 
//  through the list.  ESC and or ENTER will still exit the dialog.
//
  SetFocus (GetDlgItem(hwnd,IDC_KILIST));
  return;
}

#else WINCE_PPC

//
//  Non-PPC handler for the WM_INITDIALOG for the Kanji Info dialog.
//  
void KANJI_info::init_dialog (HWND hwnd) {
  static short types[] = { IDS_KI_TYPEUNKNOWN,IDS_KI_TYPEASCII,IDS_KI_TYPEEXTENDED,IDS_KI_TYPESYMBOL,IDS_KI_TYPEJASCII,IDS_KI_TYPEHIRAGANA,IDS_KI_TYPEKATAKANA,IDS_KI_TYPEGREEK,IDS_KI_TYPERUSSIAN,IDS_KI_TYPERESERVED,IDS_KI_TYPEKANJI1,IDS_KI_TYPEKANJI2 };
  TCHAR      buffer[SIZE_BUFFER];       // Text buffer for building lines.
  EUC_buffer line;                      // EUC line buffer used to set strings in the list-box.
  TCHAR      buffer2[20];               // Buffer for accumulating pin-yin data and ASCII value of JIS code
  byte      *ptr;                       // Pointer to current location within the data buffer.
  int        i;

  if (ch > 0x00ff) ch &= 0x7f7f;        // We accept EUC characters.
//
//  Blank out all old data.
//
#ifndef WINCE
  for (i = IDC_KITYPE; i <= IDC_KIKOREAN; i++) SetDlgItemText (hwnd,i,TEXT(""));
#else   WINCE
  for (i = IDC_KITYPE; i <= IDC_KISHKANA; i++) SetDlgItemText (hwnd,i,TEXT(""));
#endif  WINCE
  SendDlgItemMessage (hwnd,IDC_KILIST,JL_RESET,0,0);
//
//  Do JIS value
// 
//
//  Utility routine to generate the character pattern geneated by varius
//  Japanese character encodings.  This is used to see what will show 
//  up when viewing a file.
//
  if (ch <= 0x00ff) { buffer2[0] = (byte) ch; buffer2[1] = 0; }
    else { buffer2[0] = HIBYTE(ch); buffer2[1] = LOBYTE(ch); buffer2[2] = 0; }
  wsprintf (buffer,TEXT("%X (%X) [%s]"),ch,ch | 0x8080,buffer2);
  SetDlgItemText (hwnd,IDC_KIJISCODE,buffer);
  wsprintf (buffer,TEXT("%X"),jis2sjis(ch));
  SetDlgItemText (hwnd,IDC_KISHIFTJIS,buffer);
  wsprintf (buffer,TEXT("%X"),jis2unicode(ch));
  SetDlgItemText (hwnd,IDC_KIUNICODE,buffer);
  SendDlgItemMessage (hwnd,IDC_KIBIGKANJI     ,WMU_SETWINDOWVALUE,0,ch);
  SendDlgItemMessage (hwnd,IDC_KIBUSHUCHAR    ,WMU_SETWINDOWVALUE,0,0);
  SendDlgItemMessage (hwnd,IDC_KICLASSICALCHAR,WMU_SETWINDOWVALUE,0,0);
//
//  Do character type.
//
  i = HIBYTE(ch);
  if       (ch == 0)                   i = CHARTYPE_UNKNOWN;
  else if  (ch <= 127)                 i = CHARTYPE_ASCII;
  else if  (ch <= 255)                 i = CHARTYPE_OEM;
  else if ((i == 0x21) || (i == 0x22)) i = CHARTYPE_JSYMBOL;
  else if  (i == 0x23)                 i = CHARTYPE_JASCII;
  else if  (i == 0x24)                 i = CHARTYPE_HIRAGANA;
  else if  (i == 0x25)                 i = CHARTYPE_KATAKANA;
  else if  (i == 0x26)                 i = CHARTYPE_GREEK;
  else if  (i == 0x27)                 i = CHARTYPE_RUSSIAN;
  else if ((i >= 0x28) && (i <= 0x2f)) i = CHARTYPE_RESERVED;
  else if ((i >= 0x30) && (i <= 0x4f)) i = CHARTYPE_KANJI1;
  else if ((i >= 0x50) && (i <= 0x74)) i = CHARTYPE_KANJI2;
  else                                 i = CHARTYPE_UNKNOWN;
  SetDlgItemText (hwnd,IDC_KITYPE,get_string(types[i]));
#ifndef FIRST_PAGE_ONLY
  EnableWindow(GetDlgItem(hwnd,IDC_KIMORE),i >= CHARTYPE_KANJI1);
#endif  FIRST_PAGE_ONLY
  if (i < CHARTYPE_KANJI1) return;          // Not kanji we have have no more info.
  memset (&kinfo,0,sizeof(kinfo));
  if (((ch & 0xff) < 0x21) || ((ch & 0xff) > 0x7e)) return;
//
//  Open and read kanji information, an display the easy items.
//        
  if (open_info()) return;
  get_info (ch,INFO_ALL);
  close_info ();
  SendDlgItemMessage (hwnd,IDC_KIBUSHUCHAR,WMU_SETWINDOWVALUE,0,kinfo.bushu);
  SetDlgItemInt (hwnd,IDC_KISTROKES,kinfo.strokes,true);
  if (!kinfo.classical) SetDlgItemInt (hwnd,IDC_KIBUSHU,kinfo.bushu,true);
    else {
      wsprintf (buffer2,TEXT("%d (%d)"),kinfo.bushu,kinfo.classical);
      SetDlgItemText     (hwnd,IDC_KIBUSHU,buffer2);
      SendDlgItemMessage (hwnd,IDC_KICLASSICALCHAR,WMU_SETWINDOWVALUE,0,kinfo.classical);
    }       
  if (kinfo.grade  ) SetDlgItemInt  (hwnd,IDC_KIGRADE  ,kinfo.grade  ,false);
  if (kinfo.nelson ) SetDlgItemInt  (hwnd,IDC_KINELSON ,kinfo.nelson ,false);
  if (kinfo.halpern) SetDlgItemInt  (hwnd,IDC_KIHALPERN,kinfo.halpern,false);
  if (kinfo.haig   ) SetDlgItemInt  (hwnd,IDC_KIHAIG   ,kinfo.haig   ,false);
  wsprintf (buffer2,TEXT("%d-%d-%d"),kinfo.skip_t,kinfo.skip_1,kinfo.skip_2);
  if (kinfo.skip_t ) SetDlgItemText (hwnd,IDC_KISKIP,buffer2);
//
//  Process PinYin.
//
#ifndef WINCE
  if (korean) info_string (hwnd,IDC_KIKOREAN,korean);
  if (pinyin) info_string (hwnd,IDC_KIPINYIN,pinyin);
#endif WINCE
//
//  Read all of the line type entries all in one big block.  This is 
//  very brut-force, but it does save time/space, and everything else.
//
  line.initialize (GetDlgItem(hwnd,IDC_KILIST));    // Initialize buffer for output.
//
//  Process list elements.
//
  if (jfc_config.cfg.info_titles) put_line (&line,0,(byte *) "\x01\x1fmeanings\x1f",false);
  for (ptr = imi, i = 0; i < kinfo.imi; i++) ptr = put_reading(&line,0,ptr,i,kinfo.imi);
  if (jfc_config.cfg.info_titles && kinfo.on ) put_line (&line,0,(byte *) "\x01\x1fon-yomi\x1f",false);
  for (ptr = on,  i = 0; i < kinfo.on; i++) ptr = put_reading(&line,BASE_KATAKANA,ptr,i,kinfo.on);
  if (jfc_config.cfg.info_titles && kinfo.kun) put_line (&line,0,(byte *) "\x01\x1fkun-yomi\x1f",false);
  for (ptr = kun, i = 0; i < kinfo.kun; i++) ptr = put_reading(&line,BASE_HIRAGANA,ptr,i,kinfo.kun);
  if (jfc_config.cfg.info_titles && kinfo.nan) put_line (&line,0,(byte *) "\x01\x1fnanori\x1f",false);
  for (ptr = nan, i = 0; i < kinfo.nan; i++) ptr = put_reading(&line,BASE_HIRAGANA,ptr,i,kinfo.nan);
//
//  Process the entended data entries in the list.
//
  if (kinfo.extra) {
    if (extend.sh_rstroke || extend.sh_ostroke) {
      wsprintf (buffer,TEXT("%d%c%d.%d"),extend.sh_rstroke,extend.sh_radical+'a',extend.sh_ostroke,extend.sh_index);
      SetDlgItemText (hwnd,IDC_KISHDICT,buffer);
    }
    if (sh_kana) SetDlgItemInt (hwnd,IDC_KISHKANA,sh_kana,false);
#ifndef WINCE
    info_fourcorner (hwnd,IDC_KIFC ,extend.fc_main,extend.fc_index);
    info_fourcorner (hwnd,IDC_KIFC2,fc_main2      ,extend.fc_index2);
    if (extend.md_long) {
      if      (extend.md_x) wsprintf (buffer,TEXT("%dX"),extend.md_long);
      else if (extend.md_p) wsprintf (buffer,TEXT("%dP"),extend.md_long);
      else                  wsprintf (buffer,TEXT("%d") ,extend.md_long);
      SetDlgItemText (hwnd,IDC_KIMDLONG,buffer);
    }
    if (extend.md_short1) {
      wsprintf (buffer,TEXT("%d.%04d"),extend.md_short1,extend.md_short2);
      SetDlgItemText (hwnd,IDC_KIMDSHORT,buffer);
    }
#endif WINCE
  }
//
//  Activate the list so the user can use the cursor keys to move 
//  through the list.  ESC and or ENTER will still exit the dialog.
//
  SetFocus (GetDlgItem(hwnd,IDC_KILIST));
  return;
}

#endif WINCE_PPC

//
//  Open the kanji-info file.
//
//  The file is open, and a number of spot checks are preformed to make
//  sure that this is the correct version of the file.
//
//      hwnd -- Window of the caller (used to generate error messages).
//
//      RETURN -- Non-zero return indicates an error in opening the file.
//  
int KANJI_info::open_info () {
  unsigned long done,magic,size;
//
//  If we have a cached information file we do not need to open it.
//
  if (info_cache) return (false);
//
//  Open the file
//
  handle = jfc_config.open(NAME_KANJIINFO,OPEN_READ);       // Open file and check.
  if (INVALID_HANDLE_VALUE == handle) {
    ErrorMessage (true,IDS_INFO_OPENERROR,jfc_config.name());
    return       (true);
  }
//
//  If we are going to cache the file then read the file.
//
  if (jfc_config.cfg.cache_info && main_window) {           // hwnd check is to prevent load on intialization.
    size = GetFileSize(handle,&done);
    if (!(info_cache = (byte *) malloc(size))) {
      OutOfMemory ();
      close_info  ();
      return      (true);
    }
    ReadFile (handle,info_cache,size,&done,NULL);
    magic = *((long *) info_cache);
  }
//
//  Not cached then check the file information.
//
  else {
    ReadFile (handle,&magic,sizeof(long),&done,NULL);         // Check for database id (magic)
    if (magic != KINFO_MAGIC) {
      close_info ();
      handle = INVALID_HANDLE_VALUE;
      ErrorMessage (true,IDS_INFO_VERSIONERROR,jfc_config.name());
      return       (true);
    }
  }
  return (false);
}

//
//  This rutine gets all the data associated with the current kanji and stores it in
//  strings in the card data structrue.  This shield from private access to the KANJI_info
//  structrues.
//
//      ascii -- Will contain the ascii (meaning).
//      kanji -- Will contain the on-yomi.
//      kana  -- Will contian the kun-yomi.
//
void KANJI_info::put_data (KANJI *ascii,KANJI *kanji,KANJI *kana) {
  int   i;
  byte *ptr;
  for (ascii[0] = i = 0, ptr = imi; i < kinfo.imi; i++) ptr = put_entry(ascii,0            ,ptr);
  for (kanji[0] = i = 0, ptr = on ; i < kinfo.on ; i++) ptr = put_entry(kanji,BASE_KATAKANA,ptr);
  for (kana [0] = i = 0, ptr = kun; i < kinfo.kun; i++) ptr = put_entry(kana ,BASE_HIRAGANA,ptr);
  return;
}

//
//  End Class KANJI_info.
//
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//
//  Exported routines.
//

//
//  Deallocate memory resource associated with kanji information.
//
void free_info () {
  if (info_cache) free (info_cache);
  info_cache = NULL;
  return;
}

//
//  Initialies the windows classes used by the character information dialog.
//
//      wclass -- Class structure with base values initialized.
//
//      RETURN -- A non-zero return indicates a fatal error.
//
int initialize_info (WNDCLASS *wclass) {
//
//  Register window classes
//
  if (wclass) {
    wclass->style         = CS_HREDRAW | CS_VREDRAW;    // Bushu character window
    wclass->hbrBackground = (HBRUSH) (COLOR_MENU+1);
    wclass->lpfnWndProc   = JFC_bushu_proc;
    wclass->lpszClassName = TEXT("JFC-Bushu");
    if (!RegisterClass(wclass)) return (true);
    wclass->hbrBackground = (HBRUSH) (COLOR_WINDOW+1);  // Large kanji window.
    wclass->lpfnWndProc   = JFC_kanji_proc;
    wclass->lpszClassName = TEXT("JFC-Kanji");
    if (!RegisterClass(wclass)) return (true);
  }
  return (false);
}

