//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000,2001.               //
//  All rights reserved.                                             //
//                                                                   //   
//  The code do do conversion between ECU, JIS, and Shift-JIS        //
//  was taken from jconv.c which is copyright by Ken R. Lunde,       //
//  Adobe Systems Incorporated.  Full copyright notice is below.     //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------
//
//  This module implements the translatin between JWP's internal coding
//  and various other formats.  These formats include various JIS formats,
//  Shift-JIS, EUC, UNICODE, and ASCII.  (ASCII format is only supported
//  as an import format.)
//
//  This module only contains a few routines from the jfc_jisc rotuine.  
//  These are placed in the same file name to make the organization easier.
//
#include "jfc_main.h"
#include "jfc_jisc.h"

//-------------------------------------------------------------------
//
//  Static tables
//
//  These are primarally conversion talbes for processing Unicode data.
//
#define NUMBER_KANJIUNICODE     (sizeof(kanji_unicode)/sizeof(KANJI))

static KANJI kanji_unicode[] = {    // Main kanji table.  These are unicode values
  #include "jwp_ukan.dat"           //   associated with jfc_info type indexes.
};

#define NUMBER_MISCUNICODE      (sizeof(misc_unicode)/sizeof(KANJI))

static KANJI misc_unicode[] = {     // Main symbol group, corresponds to JIS codes 0x2100
  #include "jwp_umis.dat"
};

static KANJI cp1250[128] = {        // Eastern Europe code page
  #include "jwp_cp1250.dat"
};

static KANJI cp1251[128] = {        // Cyrillic code page
  #include "jwp_cp1251.dat"
};

static KANJI cp1252[128] = {        // USA, West Europe code page
  #include "jwp_cp1252.dat"
};

static KANJI cp1253[128] = {        // Greek code page
  #include "jwp_cp1253.dat"
};

static KANJI cp1254[128] = {        // Turkish code page
  #include "jwp_cp1254.dat"
};

static KANJI cp1255[128] = {        // Hebrew code page
  #include "jwp_cp1255.dat"
};

static KANJI cp1256[128] = {        // Arabic code page
  #include "jwp_cp1256.dat"
};

static KANJI cp1257[128] = {        // Baltic code page
  #include "jwp_cp1257.dat"
};

static KANJI cp1258[128] = {        // Vietnamese code page
  #include "jwp_cp1258.dat"
};

#define NUMBER_EXTUNICODE       128

static KANJI *ext_unicode = cp1252;     // Extended ascii unicode for European character support

//-------------------------------------------------------------------
//
//  Exported functions.
//

//
//  The main function of this routine is to map the UNICODE table used for extneded 
//  character processing into the table used for the local Code page.  
//
void initialize_cp () {
  switch (GetACP()) {
    case 1250: ext_unicode = cp1250; break;
    case 1251: ext_unicode = cp1251; break;
    case 1253: ext_unicode = cp1253; break;
    case 1254: ext_unicode = cp1254; break;
    case 1255: ext_unicode = cp1255; break;
    case 1256: ext_unicode = cp1256; break;
    case 1257: ext_unicode = cp1257; break;
    case 1258: ext_unicode = cp1258; break;
    default:   ext_unicode = cp1252; break;     // If we don't know the code page default to USA (my code page!)
  }
  return;
}

//
//  Attempt to see if the item connected to the buffer is in UTF-8 format.
//
//      buffer -- Pointer to buffer to be processed.
//      size   -- size of buffer
//
//      RETURN -- A zero value indicates this is UTF-8 data.
//
//  Because JIS_EOF is -1, this will cause an EOF condition to fail the tests, since all 
//  of the UTF-8 codes have at least one zero in them.  This checker is nice in that it 
//  can check just the format without having to look at the actual unicode data.
//
#define IS_UTF8(ch)     (0x80 == (ch & 0xc0))   // Test for valid second or third UTF-8 byte

int is_utf8 (byte *buffer,int size) {
  int i,j,k;
  for (i = 0; i < size; i++) {
    if      (UTF8_VALUE1 == (buffer[i] & UTF8_MASK1)) j = 0;    // Determine number of bytes we have.
    else if (UTF8_VALUE2 == (buffer[i] & UTF8_MASK2)) j = 1;
    else if (UTF8_VALUE3 == (buffer[i] & UTF8_MASK3)) j = 2;
    else if (UTF8_VALUE4 == (buffer[i] & UTF8_MASK4)) j = 3;
    else return (false);                                        // This must be smething like 1111 (not valid).
    for (k = 0; k < j; k++) {                                   // Check following bytes for correct format.
      if (++i >= size) return (false);
      if (!IS_UTF8(buffer[i])) return (false);
    }
  }
  return (true);
}

//
//  Translates from JIS code to Shift-JIS code.  This is used by the 
//  shift-JIS file writer and by the Character info screen.
//
//      ch     -- Imput character code (JIS code).
//      
//      RETURN -- Shift JIS code.
//
int jis2sjis (int ch) {
  unsigned c1,c2;
  c1  = (ch >> 8);
  c2  = (ch & 0xff);
  c2 += (c1 % 2) ? ((c2 > 95) ? 32 : 31) : 126;    // jis code -> sjis code [jis2sjis() from jconv.c]
  c1  = ((c1+1) >> 1) + ((c1 < 95) ? 112 : 176);   // jis code -> sjis code [jis2sjis() from jconv.c]
  return ((c1 << 8) | (c2));
}

//
//  This routine converts from JIS to Unicode codes.  This is one half
//  of the core of the Unicode support.
//
//      ch     -- Unicode character.
//
//      RETURN -- JIS character code.
//
int jis2unicode (int ch) {
  if ((ch <= 0x7e)                    ) return (ch);                    // ASCII
  if ((ch >= 0x0080) && (ch <= 0x00ff)) return (ext_unicode[ch-0x80]);  // Extended ASCII
  if ((ch >= 0x2330) && (ch <= 0x237a)) return (ch-0x2330+0xff10);      // Japanese ASCII
  if ((ch >= 0x2421) && (ch <= 0x2473)) return (ch-0x2421+0x3041);      // Hiragana
  if ((ch >= 0x2521) && (ch <= 0x2576)) return (ch-0x2521+0x30a1);      // Katakana
  if ((ch >= 0x2621) && (ch <= 0x2658)) return (ch-0x2621+0x0391);      // Greek
  if ((ch >= 0x2721) && (ch <= 0x2771)) return (ch-0x2721+0x0410);      // Cyrillic
  if ((ch >= 0x3021) && (ch <= 0x7426)) {                               // Kanji    
    ch -= 0x3021;
    ch  = HIBYTE(ch)*94+LOBYTE(ch);
    return (kanji_unicode[ch]);
  }
  if ((ch >= 0x2121) && (ch <= 0x217e)) return (misc_unicode[ch-0x2121]);       // Main symbol group
  if ((ch >= 0x2221) && (ch <= 0x227f)) return (misc_unicode[ch-0x2221+94]);    // Secondary symbol group
  return (0);
}

//
//  This routine converts from Unicode to JIS codes.  This is one half
//  of the core of the Unicode support.
//
//      ch     -- Unicode character.
//
//      RETURN -- JIS character code.
//
int unicode2jis (int ch) {
  int i;
  if ((ch <= 0x007e)                  ) return (ch);                // ASCII
  if ((ch >= 0x3041) && (ch <= 0x3093)) return (ch-0x3041+0x2421);  // Hiragana
  if ((ch >= 0x30a1) && (ch <= 0x30f6)) return (ch-0x30a1+0x2521);  // Katakana
  if ((ch >= 0x0391) && (ch <= 0x03c9)) return (ch-0x0391+0x2621);  // Greek
  if ((ch >= 0x0410) && (ch <= 0x044f)) return (ch-0x0410+0x2721);  // Cyrillic
  for (i = 0; i < NUMBER_KANJIUNICODE; i++) {                       // The kanji
    if (kanji_unicode[i] == ch) {
      ch = i/94;
      return (((ch+0x30) << 8) | (i-ch*94)+0x21);
    }
  }
  for (i = 0; i < NUMBER_MISCUNICODE; i++) {                        // Symbol and pucntuation (main group)
    if (misc_unicode[i] == ch) return ((i >= 94) ? 0x2221+i-94 : 0x2121+i);
  }
  for (i = 0; i < NUMBER_EXTUNICODE; i++) {                         // Extneded ascii codes.
    if (ext_unicode[i] == ch) return (i+0x80);
  }
  if ((ch >= 0xff10) && (ch <= 0xff5a)) return (ch-0xff10+0x2330);  // Japanese ASCII.
  return (0);           // The Japanese ASCII needs to be here because some puncutation is stored within gaps
}                       //   in the JASCII table, and if we move this before the misc codes they will not be found!    

